﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

using OpenGL4NET;
using System.Drawing;
using System.Windows.Forms;

namespace Semestralka
{
    class Window : Form
    {
        private Boolean fullScrean = true;

        private RenderingContext rendContext;
        private Planet sun;
        //private Planet moon;
        private Ground ground;
        float[][] pole;
        int size;
        bool right, left, forward, backward, invertedMouse, showFaces;

        Time time;

        uint[] texture = new uint[2];

        float vyska;
        float velikostPostavy = 1.85f;
        bool walk = true;
        float step = 1f;

        private double rad = Math.PI / 180;
        private double pi2 = Math.PI / 2;

        private float rychlostChuze = 3; // m/s
        private float sunAngle = 0.0f; // uhel slunce °, 0° = poledne
        private float sunR = 300f; // polomer obehu
        private float dobaObehu = 120; // s
        private float uhelSekunda;
        private long casAktualni;
        private long casPredchozi;
        private long casPredSekundou;
        private long mezicasTick;
        private float mezicasSec;
        private const float sekunda = 10000000;
        private int fps = 0;

        private float azimuth = 0;
        private float zenith = 0;
        // pohled mysi v dany cas
        private float azimuth_onTime = 0;
        private float zenith_onTime = 0;
        // vychozi pozice
        private float center = -128f;
        private float posX;
        private float posY;
        private float posZ;
        // vychozi pozice v dany cas
        private float posX_onTime = 0;
        private float posY_onTime = 0;
        private float posZ_onTime = 0;
        double azRad;
        double hoRad;

        float blue;
        float light;

        public float[] LightAmbient1 = { 0.5f, 0.5f, 0.5f, 1.0f };
        public float[] LightDiffuse1 = { 1.0f, 1.0f, 1.0f, 1.0f };
        public float[] LightPosition1;

        float camAngleOfView = 45f; //natoceni kamery
        float camNear = 0.5f;
        float camFar = 500f;

        static void Main(string[] args)
        {
            Window window = new Window();

            if (!window.init())
            {
                window.destroy();
                return;
            }
            else
            {
                window.Show();
                window.Activate();
            }

            while (window.Created)
            {
                window.render();
                Application.DoEvents();
            }

            window.destroy();

        }

        public bool init()
        {
            if (fullScrean)
            {
                this.ClientSize = new Size(System.Windows.Forms.Screen.PrimaryScreen.Bounds.Width, System.Windows.Forms.Screen.PrimaryScreen.Bounds.Height);
                this.CenterToScreen();
                this.WindowState = FormWindowState.Maximized;
                this.FormBorderStyle = FormBorderStyle.None;
                this.TopMost = true;
                this.Cursor.Dispose();
                Cursor.Position = new Point(this.Width / 2, this.Height / 2);
            }
            else
            {
                this.ClientSize = new Size(600, 480);
                this.CenterToScreen();
                this.Cursor.Dispose();
                Cursor.Position = new Point(this.Width / 2, this.Height / 2);
            }

            try
            {
                // inicializace
                pole = NactiPole.nactiPole();
                size = pole.Length * 2 - 2;
                sun = new Planet(50f);
                //moon = new Planet(20f);
                ground = new Ground(pole);
                //System.Console.Out.WriteLine("zeme vytvorena");
                rendContext = RenderingContext.CreateContext(this);
                time = new Time();
                posX = posZ = center;
                posY = 0f;
                invertedMouse = false;
                showFaces = true;
                uhelSekunda = 360 / dobaObehu;

                // inicializace GL
                gl.ShadeModel(GL.SMOOTH);
                gl.ClearColor(0, 0, 0, 0);
                gl.Clear(GL.COLOR_BUFFER_BIT | GL.DEPTH_BUFFER_BIT);

                // zapnuti pocitani hloubky
                gl.ClearDepth(1.0f);
                gl.Enable(GL.DEPTH_TEST);
                gl.DepthFunc(GL.LEQUAL);

                // nastaveni svetel
                gl.Lightfv(GL.LIGHT1, GL.AMBIENT, LightAmbient1);
                LightPosition1 = new float[] { -center, sunR, -center, 0f };
                gl.Enable(GL.LIGHT1);
                gl.Enable(GL.LIGHTING);

                // nastaveni zobrazeni a pohledove pyramidy
                gl.Disable(GL.CULL_FACE);
                gl.MatrixMode(GL.PROJECTION);
                gl.LoadMatrixf(Matrix.Perspective(camAngleOfView,
                    (float)ClientSize.Width / (float)ClientSize.Height, camNear, camFar));

                // nacteni textury
                gl.Enable(GL.TEXTURE_2D);
                gl.GenTextures(this.texture.Length, this.texture);
                LoadTextures(1, "textura.bmp");
                LoadTextures(2, "sun.bmp");
                //LoadTextures(3, "moon.bmp");

                casPredchozi = DateTime.Now.Ticks;

                return true;
            }
            catch (Exception e)
            {
                Console.WriteLine(e.Message);
                return false;
            }
        }

        private bool LoadTextures(int position, string textureName)
        {
            Bitmap image = null;
            try
            {
                image = new Bitmap(textureName);
            }
            catch (System.ArgumentException)
            {
                System.Console.Out.WriteLine("Textura nenalezena!");
            }
            if (image != null)
            {
                image.RotateFlip(RotateFlipType.RotateNoneFlipY);
                System.Drawing.Imaging.BitmapData bitmapdata;
                Rectangle rect = new Rectangle(0, 0, image.Width, image.Height);

                bitmapdata = image.LockBits(rect, System.Drawing.Imaging.ImageLockMode.ReadOnly, System.Drawing.Imaging.PixelFormat.Format24bppRgb);

                gl.BindTexture(GL.TEXTURE_2D, this.texture[position - 1]);
                gl.TexParameteri(GL.TEXTURE_2D, GL.TEXTURE_MAG_FILTER, GL.LINEAR);
                gl.TexParameteri(GL.TEXTURE_2D, GL.TEXTURE_MIN_FILTER, GL.LINEAR);
                gl.TexImage2D(GL.TEXTURE_2D, 0, (int)GL.RGB, image.Width, image.Height, 0, GL.BGR_EXT, GL.UNSIGNED_BYTE, bitmapdata.Scan0);

                image.UnlockBits(bitmapdata);
                image.Dispose();

                return true;
            }
            return false;
        }

        public void render()
        {
            casAktualni = DateTime.Now.Ticks;

            if (90 < sunAngle && sunAngle < 270)
            {
                //gl.Lightfv(GL.LIGHT1, GL.AMBIENT, new float[] { 0f, 1f, 0f, 1f });
                gl.Lightfv(GL.LIGHT1, GL.DIFFUSE, new float[] { 0f, 0f, 0f, 1.0f });
                gl.ClearColor(0, 0, 0, 0);
            }
            else
            {
                //gl.Lightfv(GL.LIGHT1, GL.AMBIENT, LightAmbient1);
                blue = Math.Abs((float)Math.Cos(sunAngle * rad));
                light = blue;
                gl.Lightfv(GL.LIGHT1, GL.DIFFUSE, new float[] { light, light, light, 1.0f });
                gl.ClearColor(0, 0, blue, 0);
            }

            gl.Clear(GL.COLOR_BUFFER_BIT | GL.DEPTH_BUFFER_BIT);

            gl.MatrixMode(GL.MODELVIEW);
            gl.LoadIdentity();

            if (showFaces)
            {
                gl.PolygonMode(GL.FRONT_AND_BACK, GL.FILL);
            }
            else
            {
                gl.PolygonMode(GL.FRONT_AND_BACK, GL.LINE);
            }

            OnTime();

            gl.Rotatef(zenith_onTime, 1.0f, 0.0f, 0.0f);
            gl.Rotatef(azimuth_onTime, 0.0f, 1.0f, 0.0f);

            Pohyb();

            gl.Translatef(posX_onTime, posY_onTime, posZ_onTime);

            gl.PushMatrix();
            gl.Rotatef(sunAngle, 0f, 0f, 1f);
            gl.Enable(GL.LIGHTING);
            gl.Lightfv(GL.LIGHT1, GL.POSITION, LightPosition1);
            gl.PopMatrix();
            gl.BindTexture(GL.TEXTURE_2D, this.texture[0]);
            ground.render();
            gl.Disable(GL.LIGHTING);

            gl.Translatef(-center, 0, -center);
            gl.Rotatef(sunAngle, 0f, 0f, 1f);
            gl.Translatef(0, sunR, 0);
            gl.BindTexture(GL.TEXTURE_2D, this.texture[1]);
            sun.render();
            //gl.BindTexture(GL.TEXTURE_2D, this.texture[2]);
            //gl.Translatef(0, -2*sunR, 0);
            //moon.render();

            rendContext.SwapBuffers();
        }

        public void OnTime()
        {
            mezicasTick = casAktualni - casPredchozi;
            mezicasSec = mezicasTick / sekunda;

            azimuth_onTime = azimuth;
            zenith_onTime = zenith;
            posX_onTime = posX;
            posY_onTime = posY;
            posZ_onTime = posZ;
            sunAngle += mezicasSec * uhelSekunda;
            sunAngle %= 360;
            step = mezicasSec * rychlostChuze;

            fps++;

            if (casAktualni - casPredSekundou > sekunda)
            {
                // System.Console.Out.WriteLine("FPS = " + fps);
                fps = 0;
                casPredSekundou = casAktualni;
            }

            casPredchozi = casAktualni;
        }

        public void Pohyb()
        {
            azRad = azimuth * rad;
            hoRad = zenith * rad;

            if ((forward && left) || (forward && right) || (backward && left) || (backward && right))
            {
                if (forward && left)
                {
                    posX += step * (float)Math.Cos(azRad + pi2 / 2);
                    posZ += step * (float)Math.Sin(azRad + pi2 / 2);
                    posY += step * (float)Math.Sin(hoRad);
                }
                if (forward && right)
                {
                    posX -= step * (float)Math.Cos(azRad - pi2 / 2);
                    posZ -= step * (float)Math.Sin(azRad - pi2 / 2);
                    posY -= step * (float)Math.Sin(hoRad);
                }
                if (backward && left)
                {
                    posX += step * (float)Math.Cos(azRad - pi2 / 2);
                    posZ += step * (float)Math.Sin(azRad - pi2 / 2);
                    posY += step * (float)Math.Sin(hoRad);
                }
                if (backward && right)
                {
                    posX -= step * (float)Math.Cos(azRad + pi2 / 2);
                    posZ -= step * (float)Math.Sin(azRad + pi2 / 2);
                    posY -= step * (float)Math.Sin(hoRad);
                }
            }
            else
            {

                if (forward)
                {
                    posX += step * (float)Math.Cos(azRad + pi2);
                    posZ += step * (float)Math.Sin(azRad + pi2);
                    posY += step * (float)Math.Sin(hoRad);
                }
                if (backward)
                {
                    posX -= step * (float)Math.Cos(azRad + pi2);
                    posZ -= step * (float)Math.Sin(azRad + pi2);
                    posY -= step * (float)Math.Sin(hoRad);
                }
                if (left)
                {
                    posX += step * (float)Math.Cos(azRad);
                    posZ += step * (float)Math.Sin(azRad);
                }
                if (right)
                {
                    posX -= step * (float)Math.Cos(azRad);
                    posZ -= step * (float)Math.Sin(azRad);
                }
            }

            if (posX > 0)
            {
                posX = 0;
            }
            if (posX < -size + 0.01f)
            {
                posX = -size + 0.01f;
            }
            if (posZ > 0)
            {
                posZ = 0;
            }
            if (posZ < -size + 0.01f)
            {
                posZ = -size + 0.01f;
            }

            if (walk)
            {
                int x2 = getPositionInField(posX);
                int z2 = getPositionInField(posZ);

                float xz = pole[z2][x2];
                float x1z = pole[z2][x2 + 1];
                float xz1 = pole[z2 + 1][x2];
                float x1z1 = pole[z2 + 1][x2 + 1];

                float xPos = -posX - 2 * x2;
                float zPos = -posZ - 2 * z2;

                vyska = xz / 4 * (2 - xPos) * (2 - zPos)
                    + x1z / 4 * (xPos) * (2 - zPos)
                    + xz1 / 4 * (2 - xPos) * (zPos)
                    + x1z1 / 4 * (xPos) * (zPos);
                posY = -vyska - velikostPostavy;

                /*
                System.Console.Out.WriteLine("-------------------------------------------------------------------");
                System.Console.Out.WriteLine("posX = " + posX + " ; posZ = " + posZ);
                System.Console.Out.WriteLine("x2 = " + x2 + " ; z2 = " + z2);
                System.Console.Out.WriteLine("xz = " + xz + " ; x1z = " + x1z + " ; xz1 = " + xz1 + " ; x1z1 = " + x1z1);
                System.Console.Out.WriteLine("xPos = " + xPos + " ; zPos = " + zPos);
                System.Console.Out.WriteLine("vyska = " + vyska + " ; posY = " + posY);
                */
            }

        }

        private int getPositionInField(float position)
        {
            int p = (int)(-position);
            return p % 2 == 0 ? p /= 2 : p = (int)((p - 1) / 2);
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            azimuth = ((float)e.X / this.Width) * 360;

            if (!invertedMouse)
            {
                zenith = (((float)e.Y / this.Height) * 180) - 90;
            }
            else
            {
                zenith = (((float)e.Y / this.Height) * -180) + 90;
            }


            if (e.X < 1)
            {
                Cursor.Position = new Point(this.Width - 2, e.Y);
            }
            if (e.X > this.Width - 2)
            {
                Cursor.Position = new Point(1, e.Y);
            }

        }

        protected override void OnKeyDown(KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.A:
                case Keys.Left: left = true; break;
                case Keys.D:
                case Keys.Right: right = true; break;
                case Keys.W: forward = true; break;
                case Keys.S: backward = true; break;
                case Keys.U: invertedMouse = !invertedMouse; break;
                case Keys.F: showFaces = !showFaces; break;
                case Keys.R:
                    posX = posZ = center;
                    break;
                case Keys.Escape: Close(); break;
            }
        }

        protected override void OnKeyUp(KeyEventArgs e)
        {
            switch (e.KeyCode)
            {
                case Keys.A:
                case Keys.Left: left = false; break;
                case Keys.D:
                case Keys.Right: right = false; break;
                case Keys.W: forward = false; break;
                case Keys.S: backward = false; break;
            }
        }

        public void destroy()
        {
            ground.destroy();
            sun.destroy();
        }
    }
}
